@@ -1,7 +1,7 @@
 package Complete::Fish::Gen::FromGetoptLong;
 
 our $DATE = '2016-10-21'; # DATE
-our $VERSION = '0.04'; # VERSION
+our $VERSION = '0.05'; # VERSION
 
 use 5.010001;
 use strict;
@@ -14,7 +14,21 @@ our %SPEC;
 
 require Exporter;
 our @ISA = qw(Exporter);
-our @EXPORT_OK = qw(gen_fish_complete_from_getopt_long_spec);
+our @EXPORT_OK = qw(
+                       gen_fish_complete_from_getopt_long_script
+                       gen_fish_complete_from_getopt_long_spec
+               );
+
+my %common_args = (
+    cmdname => {
+        summary => 'Command name',
+        schema => 'str*',
+    },
+    compname => {
+        summary => 'Completer name (in case different from cmdname)',
+        schema => 'str*',
+    },
+);
 
 $SPEC{gen_fish_complete_from_getopt_long_spec} = {
     v => 1.1,
@@ -31,14 +45,7 @@ _
             req => 1,
             pos => 0,
         },
-        cmdname => {
-            summary => 'Command name',
-            schema => 'str*',
-        },
-        compname => {
-            summary => 'Completer name (in case different from cmdname)',
-            schema => 'str*',
-        },
+        %common_args,
     },
     result => {
         schema => 'str*',
@@ -58,31 +65,94 @@ sub gen_fish_complete_from_getopt_long_spec {
 
     my @cmds;
     my $prefix = "complete -c ".shell_quote($cmdname);
-    push @cmds, "$prefix -e"; # currently does not work (fish bug)
-    for my $ospec (sort keys %$gospec) {
+    my $a_val  = shell_quote("(begin; set -lx COMP_SHELL bash; set -lx COMP_LINE (commandline); set -lx COMP_POINT (commandline -C); ".shell_quote($compname)."; end)");
+    push @cmds, "$prefix -e"; # currently does not work (fish bug?)
+    for my $ospec (sort {
+        # make sure <> is the last
+        my $a_is_diamond = $a eq '<>' ? 1:0;
+        my $b_is_diamond = $b eq '<>' ? 1:0;
+        $a_is_diamond <=> $b_is_diamond || $a cmp $b
+    } keys %$gospec) {
         my $res = parse_getopt_long_opt_spec($ospec)
             or die "Can't parse option spec '$ospec'";
-        $res->{min_vals} //= $res->{type} ? 1 : 0;
-        $res->{max_vals} //= $res->{type} || $res->{opttype} ? 1:0;
-        for my $o0 (@{ $res->{opts} }) {
-            my @o = $res->{is_neg} && length($o0) > 1 ?
-                ($o0, "no$o0", "no-$o0") : ($o0);
-            for my $o (@o) {
-                my $cmd = $prefix;
-                $cmd .= length($o) > 1 ? " -l '$o'" : " -s '$o'";
-                # XXX where to get summary from?
-                if ($res->{min_vals} > 0) {
-                    $cmd .= " -r -f -a ".shell_quote("(begin; set -lx COMP_SHELL bash; set -lx COMP_LINE (commandline); set -lx COMP_POINT (commandline -C); ".shell_quote($compname)."; end)");
+        if ($res->{is_arg}) {
+            push @cmds, "$prefix -a $a_val";
+        } else {
+            $res->{min_vals} //= $res->{type} ? 1 : 0;
+            $res->{max_vals} //= $res->{type} || $res->{opttype} ? 1:0;
+            for my $o0 (@{ $res->{opts} }) {
+                my @o = $res->{is_neg} && length($o0) > 1 ?
+                    ($o0, "no$o0", "no-$o0") : ($o0);
+                for my $o (@o) {
+                    my $cmd = $prefix;
+                    $cmd .= length($o) > 1 ? " -l '$o'" : " -s '$o'";
+                    # XXX where to get summary from?
+                    if ($res->{min_vals} > 0) {
+                        $cmd .= " -r -f -a $a_val";
+                    }
+                    push @cmds, $cmd;
                 }
-                push @cmds, $cmd;
             }
         }
     }
     [200, "OK", join("", map {"$_\n"} @cmds)];
 }
 
+$SPEC{gen_fish_complete_from_getopt_long_script} = {
+    v => 1.1,
+    summary => 'Generate fish completion script from Getopt::Long script',
+    description => <<'_',
+
+
+_
+    args => {
+        filename => {
+            schema => 'filename*',
+            req => 1,
+            pos => 0,
+        },
+        %common_args,
+    },
+    result => {
+        schema => 'str*',
+        summary => 'A script that can be fed to the fish shell',
+    },
+};
+sub gen_fish_complete_from_getopt_long_script {
+    my %args = @_;
+
+    my $filename = $args{filename};
+    return [404, "No such file or not a file: $filename"] unless -f $filename;
+
+    require Getopt::Long::Dump;
+    my $dump_res = Getopt::Long::Dump::dump_getopt_long_script(
+        filename => $filename,
+    );
+    return $dump_res unless $dump_res->[0] == 200;
+
+    my $cmdname = $args{cmdname};
+    if (!$cmdname) {
+        ($cmdname = $filename) =~ s!.+/!!;
+    }
+    my $compname = $args{compname} // $cmdname;
+
+    my $glspec = $dump_res->[2];
+
+    # GL:Complete scripts can also complete arguments
+    my $mod = $dump_res->[3]{'func.detect_res'}[3]{'func.module'} // '';
+    if ($mod eq 'Getopt::Long::Complete') {
+        $glspec->{'<>'} = sub {};
+    }
+
+    gen_fish_complete_from_getopt_long_spec(
+        spec => $dump_res->[2],
+        cmdname => $cmdname,
+        compname => $compname,
+    );
+}
+
 1;
-# ABSTRACT: From Getopt::Long spec, generate tab completion commands for the fish shell
+# ABSTRACT: Generate fish completion script from Getopt::Long spec/script
 
 __END__
 
@@ -92,17 +162,51 @@ __END__
 
 =head1 NAME
 
-Complete::Fish::Gen::FromGetoptLong - From Getopt::Long spec, generate tab completion commands for the fish shell
+Complete::Fish::Gen::FromGetoptLong - Generate fish completion script from Getopt::Long spec/script
 
 =head1 VERSION
 
-This document describes version 0.04 of Complete::Fish::Gen::FromGetoptLong (from Perl distribution Complete-Fish-Gen-FromGetoptLong), released on 2016-10-21.
+This document describes version 0.05 of Complete::Fish::Gen::FromGetoptLong (from Perl distribution Complete-Fish-Gen-FromGetoptLong), released on 2016-10-21.
 
 =head1 SYNOPSIS
 
 =head1 FUNCTIONS
 
 
+=head2 gen_fish_complete_from_getopt_long_script(%args) -> [status, msg, result, meta]
+
+Generate fish completion script from Getopt::Long script.
+
+This function is not exported by default, but exportable.
+
+Arguments ('*' denotes required arguments):
+
+=over 4
+
+=item * B<cmdname> => I<str>
+
+Command name.
+
+=item * B<compname> => I<str>
+
+Completer name (in case different from cmdname).
+
+=item * B<filename>* => I<filename>
+
+=back
+
+Returns an enveloped result (an array).
+
+First element (status) is an integer containing HTTP status code
+(200 means OK, 4xx caller error, 5xx function error). Second element
+(msg) is a string containing error message, or 'OK' if status is
+200. Third element (result) is optional, the actual result. Fourth
+element (meta) is called result metadata and is optional, a hash
+that contains extra information.
+
+Return value: A script that can be fed to the fish shell (str)
+
+
 =head2 gen_fish_complete_from_getopt_long_spec(%args) -> [status, msg, result, meta]
 
 From Getopt::Long spec, generate tab completion commands for the fish shell.
@@ -156,10 +260,6 @@ feature.
 
 =head1 SEE ALSO
 
-This module is used by L<Getopt::Long::Complete>.
-
-L<Perinci::Sub::To::FishComplete>
-
 =head1 AUTHOR
 
 perlancar <perlancar@cpan.org>
